from .datastorage import DataStorage
from .utils import get_dataset_property

class Filesystem(DataStorage):

    # overriding abstract method
    def load_dataset(self, name, read_all=False, input_dest='', header=True):

        if input_dest == '':
            input_dataset = get_dataset_property(name)
        else:
            input_dataset = input_dest
        
        delimiter = get_dataset_property(name, "delimiter") if get_dataset_property(name, "delimiter") is not None else ","
        input_file_format = get_dataset_property(name, "inputFileFormat") if get_dataset_property(name, "inputFileFormat") is not None else "csv"
        
        df = self.spark.read.load(input_dataset, format=input_file_format, sep=delimiter, inferSchema=True, header=header, error_bad_lines=False)

        if read_all == False:
            input_cols = get_dataset_property(name, "input_columns")
            if input_cols is not None and input_cols != "*":
                df = df.select(input_cols.split(delimiter))

        return df

    # overriding abstract method
    def save_dataset(self, name, df, output_dest):

        if output_dest == '':
            output_dataset = get_dataset_property(name)
        else:
            output_dataset = output_dest
        delimiter = get_dataset_property(name, "delimiter") if get_dataset_property(name, "delimiter") is not None else ","
        output_file_format = get_dataset_property(name, "outputFileFormat") if get_dataset_property(name, "outputFileFormat") is not None else "csv"

        if get_dataset_property(name, "save_dataset_with_pandas") is not None:
            df.toPandas().to_csv(output_dataset, sep=delimiter, index=False)
        else:
            df.write.options(delimiter=delimiter).\
                mode('overwrite').\
                option("header", True).\
                save(output_dataset, format=output_file_format)



from .datastorage import DataStorage
from .utils import get_dataset_property

class Presto(DataStorage):

    # Initialize Presto Python Driver

    # overriding abstract method
    def load_dataset(self, name, read_all=True, input_dest='', header=True):

        input_columns = get_dataset_property("input_columns") if get_dataset_property("input_columns") is not None else "*"

        if input_dest == '':
            input_dataset = get_dataset_property(name)
        else:
            input_dataset = input_dest

        password = get_dataset_property(name, "prestoPassword")
        user = get_dataset_property(name, "prestoUser")
        host = get_dataset_property(name, "prestoHost")
        port = get_dataset_property(name, "prestoPort")
        db = get_dataset_property(name, "prestoDb")

        if password is not None and password not in ['', "", "''"]:
            if read_all:
                df = self.spark.read.format("jdbc").option("driver", "com.facebook.presto.jdbc.PrestoDriver"). \
                    option("url", "jdbc:presto://" + host+":"+port+"/" + db).option("dbtable", "(SELECT * FROM " + input_dataset + ")"). \
                    option("user", user). \
                    option('password', password).load()
            else:
                df = self.spark.read.format("jdbc").option("driver", "com.facebook.presto.jdbc.PrestoDriver"). \
                    option("url", "jdbc:presto://"+host+":"+port+"/"+db).option("dbtable", "(SELECT " + input_columns + " FROM " + input_dataset + ")"). \
                    option("user", user). \
                    option('password', password).load()
        else:
            if read_all:
                df = self.spark.read.format("jdbc").option("driver", "com.facebook.presto.jdbc.PrestoDriver"). \
                    option("url", "jdbc:presto://"+host+":"+port+"/"+db).option("dbtable", "(SELECT * FROM " + input_dataset + ")"). \
                    option("user", get_dataset_property("prestoUser-input-dataset")).load()
            else:
                df = self.spark.read.format("jdbc").option("driver", "com.facebook.presto.jdbc.PrestoDriver"). \
                    option("url", "jdbc:presto://"+host+":"+port+"/"+db).option("dbtable", "(SELECT " + input_columns + " FROM " + input_dataset + ")"). \
                    option("user", user).load()

        # rename column
        for column in input_columns.split(","):
            df = df.withColumnRenamed(column.lower(), column)
        return df

    # overriding abstract method
    def save_dataset(self, name, df, output_dest):
        pass



from ..utils import get_model_property
from .model import ModelStorage
import os

class Minio(ModelStorage):
    storage_type = os.path.basename(__file__).split('.py')[0]

    def __init__(self, spark, name):
        self.spark = spark
        self.name = name

    def load_model(self):
        return "s3a://" + get_model_property(name=self.name, property='minio_bucket')+ "/" + get_model_property(name=self.name)

    def save_model(self, model):
        model_path = "s3a://" + get_model_property(name=self.name, property='minio_bucket')+ "/" + get_model_property(name=self.name)
        model.write().overwrite().save(model_path)





# This is a list of always applied configurations. They do not depend on the application.

starting_config = [

    # MinIO configs
    ("spark.hadoop.fs.s3a.path.style.access", True),
    ("spark.hadoop.fs.s3a.impl", "org.apache.hadoop.fs.s3a.S3AFileSystem"),

    # Hive configs
    ('hive.server2.enable.doAs', True),
    ('hive.metastore.client.connect.retry.delay', 5),
    ('hive.metastore.client.socket.timeout', 1800),
    ('spark.hadoop.dfs.client.use.datanode.hostname', True),
    ('spark.sql.warehouse.dir', '/warehouse/tablespace/managed/hive'),
    ("spark.sql.catalogImplementation", "hive"),

]


from pyspark.sql import SparkSession
from pyspark import SparkConf
from .config import starting_config

def spark_initializer(app_name, args, additional_config=[]):

    # Create list of configurations
    additional_config = starting_config + auto_conf_generator(args) + additional_config

    config = SparkConf().setAll(additional_config)

    spark = SparkSession.builder.appName(app_name).config(conf=config).getOrCreate()

    return spark



# Translate automatically configurations from services' arguments to spark properties
def auto_conf_generator(args):
    
    args_name = {
        # Hive
        "spark.sql.warehouse.dir" : "spark.sql.warehouse.dir",
        "hiveMetastoreUris" : "hive.metastore.uris",
        "hiveMetastoreUris" : "hive.metastore.uris",
        # MinIO
        "minIO_URL" : "spark.hadoop.fs.s3a.endpoint",
        "minIO_ACCESS_KEY" : "spark.hadoop.fs.s3a.access.key",
        "minIO_SECRET_KEY" : "spark.hadoop.fs.s3a.secret.key" ,
        "use_ssl" : "spark.hadoop.fs.s3a.connection.ssl.enabled"
    }

    result = []

    # for key in args_name:
    #     if key in args:
    #         result.append((args_name[key], args[key]))
    
    for key in args_name:
        for arg in args:
            if key in arg:
                result.append((args_name[key], args[arg]))
    
    return result




from ..utils import get_model_property
from .model import ModelStorage
import os

class Filesystem(ModelStorage):
    storage_type = os.path.basename(__file__).split('.py')[0]

    def __init__(self, spark, name):
        self.spark = spark
        self.name = name

    def load_model(self, name):
        return get_model_property(name=name)

    def save_model(self, model):
        model.write().overwrite().save(get_model_property(name=self.name))



from .datastorage import DataStorage

from .filesystem import Filesystem
from .hdfs import Hdfs
from .hive import Hive
from .presto import Presto

from .model import ModelStorage
from .model import minio, filesystem, hdfs


from ..utils import get_model_property
from .model import ModelStorage
import os
from bdaserviceutils import get_cmd_arg

class Hdfs(ModelStorage):
    storage_type = os.path.basename(__file__).split('.py')[0]

    def __init__(self, spark, name):
        self.spark = spark
        self.name = name

    def load_model(self):
        return get_cmd_arg('hdfsUrl') + get_model_property(name=self.name)

    def save_model(self, model):
        model_path = get_cmd_arg('hdfsUrl') + get_model_property(name=self.name)
        model.write().overwrite().save(model_path)




from abc import ABC
from ..utils import get_model_property


class ModelStorage(ABC):

    def __new__(cls, *args, **kw):
        
        storage_type = get_model_property(kw['name'], "storage_type")
        
        # Create a map of all subclasses based on storage type property (present on each subclass)
        subclass_map = {subclass.storage_type: subclass for subclass in cls.__subclasses__()}

        # Select the proper subclass based on
        subclass = subclass_map[storage_type.lower()]
        return super(cls, subclass).__new__(subclass)
    
    def __init__(self, spark, name):
        self.spark = spark
    
    # abstract method
    def load_model(self):
        pass

    # abstract method
    def save_model(self, model):
        pass
